#!/usr/bin/env bash
##
## Copyright (C) HySoP 2011-2022
##
## Licensed under the Apache License, Version 2.0 (the "License");
## you may not use this file except in compliance with the License.
## You may obtain a copy of the License at
##
##         http://www.apache.org/licenses/LICENSE-2.0
##
## Unless required by applicable law or agreed to in writing, software
## distributed under the License is distributed on an "AS IS" BASIS,
## WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
## See the License for the specific language governing permissions and
## limitations under the License.
##
set -feu -o pipefail

############################################################################
# /!\ This script will delete old CI pipelines with logs and atrifacts /!\ #
############################################################################

# This script is adapted from https://gist.github.com/mgoellnitz/45f4cbe6fa4d757a1e9803094361b846
# to conform with gitlab API v4 and requires python2.7 with the json module available

function print_help {
  echo "GitLab continuous integration pipeline cleanup 'clean_ci.sh'"
  echo ""
  echo "$0 [-u url] [-d date] [-t token] [-p project]"
  echo ""
  echo "  -u instance base url - default https://gricad-gitlab.univ-grenoble-alpes.fr"
  echo ""
  echo "  -n gitlab namespace - defaults to 'particle_methods'."
  echo ""
  echo "  -p project name - defaults to 'hysop'."
  echo ""
  echo "  -d yyyy-mm-dd - date to clean builds before - defaults to one month ago"
  echo ""
  echo "  -t gitlab api token - defaults to \$GITLAB_API_TOKEN"
  echo ""
  echo "You need a maintainer or owner gitlab token to be able to delete pipeline jobs."
  echo "Error message 'Job is not erasable!' means the job has already been erased."
  echo "Some job removal may also fail because they did not complete in the pipeline."
  exit 1
}

function extract_json {
    script=${1-"print data['id'];"}
    python2.7 -c "import sys, json, operator as op; data=json.load(sys.stdin); ${script}"
}
function api_request {
    curl -H "${CONTENT}" -H "${HEADER}" "${1}" 2>/dev/null
}

GITLAB='https://gricad-gitlab.univ-grenoble-alpes.fr'
NAMESPACE='particle_methods'
PROJECT='hysop'
DATE="$[ $(date +%s) - 2592000 ]"
TOKEN="${GITLAB_API_TOKEN:-}"

# Parse args
if [[ ${#} -gt 0 ]]; then
    PSTART=$(echo ${1} | sed -e 's/^\(.\).*/\1/g')
    while [ "$PSTART" = "-" ] ; do
      if [ "${1}" = "-u" ] ; then
        shift
        GITLAB="${1}"
      elif [ "${1}" = "-n" ] ; then
        shift
        NAMESPACE="${1}"
      elif [ "${1}" = "-p" ] ; then
        shift
        PROJECT="${1}"
      elif [ "${1}" = "-t" ] ; then
        shift
        TOKEN="${1}"
      elif [ "${1}" = "-d" ] ; then
        shift
        DATE="$(date -d ${1} +%s)"
      else
          print_help
      fi
      shift
      PSTART=$(echo ${1} | sed -e 's/^\(.\).*/\1/g')
    done
    echo "PSTART='${PSTART}'"
fi

if [[ -z "${GITLAB}" ]]; then
    echo "Gitlab url has not been specified."
    exit 1
fi
if [[ -z "${NAMESPACE}" ]]; then
    echo "Namespace has not been specified."
    exit 1
fi
if [[ -z "${PROJECT}" ]]; then
    echo "Project has not been specified."
    exit 1
fi
if [[ "${TOKEN}" == "" ]]; then
    echo "Could not determine gitlab API token, please define GITLAB_API_TOKEN or pass it explicitely."
    exit 1
fi

echo "Cleaning CI pipelines older than $(date -d @${DATE}) for project ${GITLAB}/${NAMESPACE}/${PROJECT} with token ${TOKEN}."

CONTENT='Content-Type: application/json'
HEADER="PRIVATE-TOKEN: ${TOKEN}"

PROJECT_ID="$(api_request "${GITLAB}/api/v4/projects/${NAMESPACE}%2F${PROJECT}" | extract_json)"
API_URL="${GITLAB}/api/v4/projects/${PROJECT_ID}"
if [[ -z "${PROJECT_ID}" ]]; then
    echo "Could not determine project id for project ${GITLAB}/${NAMESPACE}/${PROJECT}."
    exit 1
fi

page=1
pipelines=$(api_request "${API_URL}/pipelines?page=${page}&per_page=100" | extract_json "print('\n'.join('{}@{}'.format(d['id'], d['created_at']) for d in sorted(data, reverse=True)))")
until [[ -z "${pipelines}" ]]; do
    for pipeline in ${pipelines}; do
        pipeline_id="$(echo "${pipeline}" | cut -f1 -d'@')"
        pipeline_date="$(echo "${pipeline}" | cut -f2 -d'@' | sed -e 's/^\(.*\)T.*/\1/g')"
        PDATE="$(date -d "${pipeline_date}" +%s)"
        if [[ "${PDATE}" -ge ${DATE} ]]; then
            continue
        fi
        echo "Found pipeline ${pipeline_id} created on ${pipeline_date}:"
        set +e
        for job_id in $(api_request "${API_URL}/pipelines/${pipeline_id}/jobs" | extract_json "print '\n'.join(map(str, sorted(map(op.itemgetter('id'), data), reverse=True)))"); do
            status="$(curl -H "${CONTENT}" -H "${HEADER}" --request POST "${API_URL}/jobs/${job_id}/erase" 2>/dev/null | extract_json "print data.get('message', data.get('status', 'fail'));")"
            echo " | removing job ${job_id}... ${status}"
        done
        #status="$(curl -H "${CONTENT}" -H "${HEADER}" --request 'DELETE' "${API_URL}/pipelines/${pipeline_id}" 2>/dev/null | extract_json "print data.get('status', data.get('message', 'fail'));")"
        #echo " | removing pipeline ${pipeline_id}... ${status}"
        # Not working yet, requires to be owner ?
        set -e
    done;
    page=$((${page}+1))
    pipelines=$(api_request "${API_URL}/pipelines?page=${page}&per_page=100" | extract_json "print('\n'.join('{}@{}'.format(d['id'], d['created_at']) for d in sorted(data, reverse=True)))")
    break
done;
exit 0
