#
# Copyright (C) HySoP 2011-2022
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from hysop.problem          import Problem
from hysop.operators        import DirectionalAdvection, PoissonCurl
from hysop.numerics.splitting.strang import StrangSplitting

from hysop.domain.box         import Box
from hysop.fields.continuous_field  import Field
from hysop.topology.cartesian_topology import CartesianTopology
from hysop.tools.parameters   import CartesianDiscretization
from hysop.simulation import Simulation

from hysop.methods import *

class TestStrang:

    @classmethod
    def setup_class(cls):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def setup_method(self, method):
        pass

    def teardown_method(self, method):
        pass

    def make_fields(self, dim, nscalars):
        # Domain
        box = Box(length=[1.0]*dim, origin=[0.0]*dim)

        ## Fields
        velo = Field(domain=box, formula=None,
                     name='V', is_vector=True)
        if dim==2:
            vorti = Field(domain=box, name='W')
        else:
            vorti = Field(domain=box, formula=None,
                    name='W', is_vector=True)

        scalars = [ Field(domain=box, name=f'S{i}')
                           for i in range(nscalars)]

        return box, velo, vorti, tuple(scalars)

    def _test_strang_nd(self, dim, n):
        ## Simulation
        dt0=0.1
        simu = Simulation(start=0.0, end=1.0, max_iter=100, dt0=dt0)
        simu.initialize()

        ## Domain and continuous fields
        box, velo, vorti, scalars = self.make_fields(dim,2)

        ## CartesianDiscretizations and topologies
        resolution     = (n,)*dim
        dnd            = CartesianDiscretization(resolution=resolution, ghosts=None, default_boundaries=True)
        topo           = CartesianTopology(domain=box,discretization=dnd)

        ## Operators
        advec = DirectionalAdvection(
            name='advec',
            dt = simu.dt,
            velocity = velo,velocity_cfl=1.*dt0*n,
            advected_fields = (vorti,)+scalars,
            variables = { velo: topo, vorti: topo,
                          scalars[0]: topo, scalars[1]: topo},
            )

        poisson = PoissonCurl(name='poisson',
                              velocity=velo, vorticity=vorti,
                              variables={velo: topo, vorti: topo})

        ## Graph
        splitting = StrangSplitting(splitting_dim=dim,
                                    order=StrangOrder.STRANG_FIRST_ORDER)
        splitting.push_operators(advec)

        problem = Problem()
        problem.insert(splitting)
        problem.insert(poisson)
        problem.build()
        problem.finalize()

    def test_strang_2d(self, n=33):
        self._test_strang_nd(dim=2, n=n)
    def test_strang_3d(self, n=33):
        self._test_strang_nd(dim=3, n=n)

if __name__ == '__main__':
    test = TestStrang()
    test.setup_class()
    test.test_strang_3d()
    test.teardown_class()
